/*
 * apple2.c
 *
 * driver for Apple ][ computers
 */

/* $Id: apple2.c,v 1.5 2000/03/06 00:39:11 nyef Exp $ */

#include <stdlib.h>

#include "ui.h"
#include "cal.h"
#include "types.h"
#include "tool.h"
#include "apple2_vdp.h"


/* manifest constants */

#define APPLE2_MEMORY_SIZE 0xc000

#define APPLE2_NUM_ROMS 6
#define ROM_D0 0
#define ROM_D8 1
#define ROM_E0 2
#define ROM_E8 3
#define ROM_F0 4
#define ROM_F8 5


/* emulated system data structures */

struct apple2_mainboard {
    cal_cpu cpu;
    apple2_vdp vdp;
    u8 *memory;

    u8 *roms[APPLE2_NUM_ROMS];

    u8 keyboard_latch;
};

struct apple2_mainboard apple2_true;
struct apple2_mainboard *apple2 = &apple2_true;


/* keyboard I/O */

void apple2_press_keyboard(struct apple2_mainboard *apple2, u8 key)
{
    apple2->keyboard_latch = key | 0x80;
}

u8 apple2_read_keyboard(struct apple2_mainboard *apple2)
{
    return apple2->keyboard_latch;
}

void apple2_clear_keyboard_strobe(struct apple2_mainboard *apple2)
{
    apple2->keyboard_latch &= 0x7f;
}


/* main I/O dispatch */

u8 apple2_read_io(u32 addr)
{
/*      deb_printf("io read 0x%04x.\n", addr); */
    if (addr < 0xc080) { /* Built-In I/O */
	if (addr < 0xc010) { /* keyboard */
	    return apple2_read_keyboard(apple2);
	} else if (addr < 0xc020) { /* keyboard strobe */
	    apple2_clear_keyboard_strobe(apple2);
	    return 0x00; /* FIXME: real return value? */
	} else if (addr < 0xc050) { /* unimplemented */
	    return 0x00;
	} else if (addr < 0xc058) { /* Video */
	    a2v_io(apple2->vdp, addr);
	    return 0x00;
	}
    }
    return 0x00;
}

void apple2_write_io(u32 addr, u8 data)
{
/*      deb_printf("io write 0x%04x, 0x%02x.\n", addr, data); */
    if (addr < 0xc080) { /* Built-In I/O */
	if (addr < 0xc010) { /* keyboard, read-only (ignore) */
	} else if (addr < 0xc020) { /* keyboard strobe */
	    apple2_clear_keyboard_strobe(apple2);
	} else if (addr < 0xc050) { /* unimplemented */
	    return;
	} else if (addr < 0xc058) { /* Video */
	    a2v_io(apple2->vdp, addr);
	    return;
	}
    }
}


/* memory access emulation */

u8 apple2_read_ram(cal_cpu cpu, u32 addr)
{
    return apple2->memory[addr];
}

void apple2_write_ram(cal_cpu cpu, u32 addr, u8 data)
{
    apple2->memory[addr] = data;
}

u8 apple2_read_rom(cal_cpu cpu, u32 addr)
{
    int rom_bank;
    
    if (addr < 0xd000) {
	return apple2_read_io(addr);
    } else {
	rom_bank = (addr - 0xd000) >> 11;

	if (!apple2->roms[rom_bank]) {
	    return 0x00;
	}
	
	return apple2->roms[rom_bank][addr & 0x7ff];
    }
}

void apple2_write_rom(cal_cpu cpu, u32 addr, u8 data)
{
    if (addr < 0xd000) {
	apple2_write_io(addr, data);
    }
}

memread8_t apple2_readtable[4] = {
    apple2_read_ram,
    apple2_read_ram,
    apple2_read_ram,
    apple2_read_rom,
};

memwrite8_t apple2_writetable[4] = {
    apple2_write_ram,
    apple2_write_ram,
    apple2_write_ram,
    apple2_write_rom,
};


/* timeslicer */

void apple2_timeslice(struct apple2_mainboard *apple2)
{
    int i;

    for (i = 243; i && (system_flags == F_NONE); i--) {
	if (system_flags == F_NONE) {
	    apple2->cpu->runfor(apple2->cpu, 65);
	    apple2->cpu->run(apple2->cpu);
	    a2v_periodic(apple2->vdp);
	    if (system_flags & F_UNIMPL) {
		deb_printf("Unimplimented Instruction\n");
		system_flags |= F_QUIT;
	    }
	}
    }
}


/* system setup */

void apple2_load_rom(char *filename, int rom)
{
    rom_file romfile;
    
    deb_printf("attempting to load %02X ROM.\n", 0xd0 + (rom << 3));
    
    romfile = read_romimage(filename);

    if (!romfile) {
	deb_printf("could not open %02X ROM.\n", 0xd0 + (rom << 3));
	deb_printf("things could get ugly.\n");
	apple2->roms[rom] = NULL;
	return;
    }

    apple2->roms[rom] = romfile->data;

    /* FIXME: memory leak */
}

void apple2_load_bios(void)
{
    apple2->roms[ROM_D0] = NULL;
    apple2->roms[ROM_D8] = NULL;
    apple2->roms[ROM_E0] = NULL;
    apple2->roms[ROM_E8] = NULL;
    apple2->roms[ROM_F0] = NULL;
    apple2_load_rom("monitor_f8.rom", ROM_F8);
/*      apple2_load_rom("autostart_f8.rom", ROM_F8); */
}

void apple2_init_cpu(void)
{
    cal_cpu cpu;
    
    cpu = cal_create(CPUT_MARAT6502);
    apple2->cpu = cpu;
    cpu->setzpage(cpu, apple2->memory);
    cpu->setmmu8(cpu, 14, 3, apple2_readtable, apple2_writetable);
    cpu->reset(cpu);
}

void apple2_init_keyboard(void)
{
    extern void kb_init(void);

    kb_init();
}

void apple2_run(rom_file romfile)
{
    if (romfile) {
	deb_printf("apple2_run: ignoring romfile.\n");
    }

    apple2->memory = malloc(APPLE2_MEMORY_SIZE);

    if (!apple2->memory) {
	deb_printf("unable to allocate system memory.\n");
	return;
    }
    
    apple2_load_bios();

    apple2_init_cpu();

    apple2_init_keyboard();
    
    apple2->vdp = a2v_init(apple2->memory); /* FIXME: return value? */

    set_timeslice((void (*)(void *)) &apple2_timeslice, apple2);
}

/*
 * $Log: apple2.c,v $
 * Revision 1.5  2000/03/06 00:39:11  nyef
 * added code to initialize keyboard
 *
 * Revision 1.4  2000/03/02 03:10:40  nyef
 * added support for loading roms other than the F8 rom
 *
 * Revision 1.3  2000/03/02 02:53:28  nyef
 * added keyboard I/O emulation
 * cleaned up some random stuff
 *
 * Revision 1.2  2000/01/01 20:54:23  nyef
 * fixed to use the correct number of cycles per line
 *
 * Revision 1.1  2000/01/01 04:12:14  nyef
 * Initial revision
 *
 */
