/*
 * nes_ppu.c
 *
 * NES PPU emulation
 */

/* $Id: nes_ppu.c,v 1.58 1999/01/11 02:23:17 nyef Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "nes_ppu.h"
#include "nes.h"
#include "tool.h"
#include "ui.h"
/* #include "emu6502.h" */
#include "io_nes.h"
#include "video.h"
#include "debug.h"
#include "mappers.h"
#include "nes_psg.h"
#include "cal.h"
#include "blitters.h"

int ppu_flags;

unsigned char PPU_sprite_ram[0x100];
unsigned char PPU_palette_ram[0x20];

/* the ppu memory areas */
unsigned char *PPU_bank_00 = NULL;
unsigned char *PPU_bank_04 = NULL;
unsigned char *PPU_bank_08 = NULL;
unsigned char *PPU_bank_0c = NULL;
unsigned char *PPU_bank_10 = NULL;
unsigned char *PPU_bank_14 = NULL;
unsigned char *PPU_bank_18 = NULL;
unsigned char *PPU_bank_1c = NULL;
unsigned char *PPU_bank_20 = NULL;
unsigned char *PPU_bank_24 = NULL;
unsigned char *PPU_bank_28 = NULL;
unsigned char *PPU_bank_2c = NULL;
unsigned char *PPU_bank_30 = NULL;
unsigned char *PPU_bank_34 = NULL;
unsigned char *PPU_bank_38 = NULL;
unsigned char *PPU_bank_3c = NULL;

/* the indirection tables */
unsigned char **PPU_banks[0x10] = {
    &PPU_bank_00,
    &PPU_bank_04,
    &PPU_bank_08,
    &PPU_bank_0c,
    &PPU_bank_10,
    &PPU_bank_14,
    &PPU_bank_18,
    &PPU_bank_1c,
    &PPU_bank_20,
    &PPU_bank_24,
    &PPU_bank_28,
    &PPU_bank_2c,
    &PPU_bank_30,
    &PPU_bank_34,
    &PPU_bank_38,
    &PPU_bank_3c,
};

unsigned char ***PPU_patterns[2] = {
    &PPU_banks[0],
    &PPU_banks[4],
};

unsigned char *PPU_nametables[4];

ppulatch_t ppu_latchfunc; /* for mapper support */

unsigned char ppu_cntrl1; /* $2000 */
unsigned char ppu_cntrl2; /* $2001 */
unsigned char ppu_status; /* $2002 */
unsigned char ppu_spraddy; /* $2003 */

unsigned char ppu_vscrol; /* $2005 */
unsigned char ppu_hscrol; /* $2005 */
unsigned char ppu_scrollmode; /* $2005 */

long PPU_addr; /* $2006/$2007 */
long PPU_amode; /* $2006 */
unsigned char ppu_addrsave; /* $2006 */
unsigned char ppu_read_latch; /* $2007 */

void PPU_render_scan_line(unsigned short line);

/* phase: location of PPU refresh. */
/* 0: VBlank.  1: pre-Hit.  2: post-Hit. */
int ppu_phase;
int ppu_scanline;
unsigned char ppu_saved_vscrol;
unsigned char ppu_saved_hscrol;
unsigned char ppu_saved_nametable;

#define P_VBLANK   0
#define P_PRE_HIT  1
#define P_POST_HIT 2

void PPU_enter_cpu(cal_cpu cpu)
{
    switch (ppu_phase) {
    case P_VBLANK:
	if (cpu->timeleft(cpu) <= 0) {
	    ppu_phase = P_PRE_HIT;
	    cpu->runfor(cpu, 114);
	    ppu_scanline = 0;
	    if (ppu_cntrl2 & 0x08) {
		ppu_status = 0x10;
	    } else {
		ppu_status = 0x00;
	    }
	    ppu_saved_vscrol = ppu_vscrol;
	    ppu_saved_hscrol = ppu_hscrol;
	    ppu_saved_nametable = ppu_cntrl1 & 3;
	}
	break;
    case P_PRE_HIT: case P_POST_HIT:
	if (cpu->timeleft(cpu) <= 0) {
	    if (ppu_scanline < 240) {
		cpu->runfor(cpu, 114);
		PPU_render_scan_line(ppu_scanline);
		cur_hsync(cpu);
	    } else if (ppu_scanline > 242) {
		if (ppu_cntrl1 & 0x80) {
		    cpu->nmi(cpu);
		}
		ppu_phase = P_VBLANK;
		cpu->runfor(cpu, 2166);
		cur_hsync(cpu);
	    } else if (ppu_scanline == 240) {
		cpu->runfor(cpu, 114);
		ppu_status = 0x80;
		ppu_scrollmode = 0;
		cur_vsync(cpu);
		nes_psg_vsync();
	    } else {
		cpu->runfor(cpu, 114);
		cur_hsync(cpu);
	    }
	    ppu_scanline++;
	}
	break;
    default:
	deb_printf("PPU_enter_cpu(): invalid phase (%d), resetting to vblank\n", ppu_phase);
	ppu_phase = P_VBLANK;
	ppu_status = 0x80;
	PPU_enter_cpu(cpu);
    }
}

void PPU_leave_cpu(cal_cpu cpu)
{
    video_events();
    if ((ppu_phase == P_VBLANK) && (cpu->timeleft(cpu) <= 0)) {
	PPU_render_video_frame();
    }
}

void PPU_init(void)
{
    ppu_phase = 0;
    ppu_flags = PPUF_NTSC;
    
    PPU_memory = malloc(0x4000);
    
    PPU_bank_00 = &(PPU_memory[0x0000]);
    PPU_bank_04 = &(PPU_memory[0x0400]);
    PPU_bank_08 = &(PPU_memory[0x0800]);
    PPU_bank_0c = &(PPU_memory[0x0c00]);
    PPU_bank_10 = &(PPU_memory[0x1000]);
    PPU_bank_14 = &(PPU_memory[0x1400]);
    PPU_bank_18 = &(PPU_memory[0x1800]);
    PPU_bank_1c = &(PPU_memory[0x1c00]);
    PPU_bank_20 = &(PPU_memory[0x2000]);
    PPU_bank_24 = &(PPU_memory[0x2400]);
    PPU_bank_28 = &(PPU_memory[0x2800]);
    PPU_bank_2c = &(PPU_memory[0x2c00]);
    PPU_bank_30 = &(PPU_memory[0x3000]);
    PPU_bank_34 = &(PPU_memory[0x3400]);
    PPU_bank_38 = &(PPU_memory[0x3800]);
    PPU_bank_3c = &(PPU_memory[0x3c00]);
    if (read_byte(romimage, 6) & 1) {
	PPU_mirror_vertical();
    } else {
	PPU_mirror_horizontal();
    }
}

#define REMAP_MEM
#define PM_VERT 1
#define PM_HORZ 2
#define PM_1_LO 3
#define PM_1_HI 4

int ppu_last_mirror;

void PPU_mirror_horizontal()
{
    if (ppu_last_mirror != PM_HORZ) {
	deb_printf("ppu: mirroring horizontal.\n");
	ppu_last_mirror = PM_HORZ;
    }
#ifdef REMAP_MEM
    PPU_bank_20 = &(PPU_memory[0x2000]);
    PPU_bank_24 = &(PPU_memory[0x2000]);
    PPU_bank_28 = &(PPU_memory[0x2400]);
    PPU_bank_2c = &(PPU_memory[0x2400]);
#endif
    PPU_nametables[0] = &(PPU_memory[0x2000]);
    PPU_nametables[1] = &(PPU_memory[0x2000]);
    PPU_nametables[2] = &(PPU_memory[0x2400]);
    PPU_nametables[3] = &(PPU_memory[0x2400]);
}

void PPU_mirror_vertical()
{
    if (ppu_last_mirror != PM_VERT) {
	deb_printf("ppu: mirroring vertical.\n");
	ppu_last_mirror = PM_VERT;
    }
#ifdef REMAP_MEM
    PPU_bank_20 = &(PPU_memory[0x2000]);
    PPU_bank_24 = &(PPU_memory[0x2400]);
    PPU_bank_28 = &(PPU_memory[0x2000]);
    PPU_bank_2c = &(PPU_memory[0x2400]);
#endif
    PPU_nametables[0] = &(PPU_memory[0x2000]);
    PPU_nametables[1] = &(PPU_memory[0x2400]);
    PPU_nametables[2] = &(PPU_memory[0x2000]);
    PPU_nametables[3] = &(PPU_memory[0x2400]);
}

void PPU_mirror_one_low()
{
    if (ppu_last_mirror != PM_1_LO) {
	deb_printf("ppu: mirroring from $2000.\n");
	ppu_last_mirror = PM_1_LO;
    }
#ifdef REMAP_MEM
    PPU_bank_20 = &(PPU_memory[0x2000]);
    PPU_bank_24 = &(PPU_memory[0x2000]);
    PPU_bank_28 = &(PPU_memory[0x2000]);
    PPU_bank_2c = &(PPU_memory[0x2000]);
#endif
    PPU_nametables[0] = &(PPU_memory[0x2000]);
    PPU_nametables[1] = &(PPU_memory[0x2000]);
    PPU_nametables[2] = &(PPU_memory[0x2000]);
    PPU_nametables[3] = &(PPU_memory[0x2000]);
}

void PPU_mirror_one_high()
{
    if (ppu_last_mirror != PM_1_HI) {
	deb_printf("ppu: mirroring from $2400.\n");
	ppu_last_mirror = PM_1_HI;
    }
#ifdef REMAP_MEM
    PPU_bank_20 = &(PPU_memory[0x2400]);
    PPU_bank_24 = &(PPU_memory[0x2400]);
    PPU_bank_28 = &(PPU_memory[0x2400]);
    PPU_bank_2c = &(PPU_memory[0x2400]);
#endif
    PPU_nametables[0] = &(PPU_memory[0x2400]);
    PPU_nametables[1] = &(PPU_memory[0x2400]);
    PPU_nametables[2] = &(PPU_memory[0x2400]);
    PPU_nametables[3] = &(PPU_memory[0x2400]);
}

void PPU_write(unsigned short address, unsigned char value)
{
    address = address & 0x3fff;
    
/*     deb_printf("PPU_write(%04hx, %02hx)\n", address, value); */

    if (ppu_latchfunc) {
	ppu_latchfunc(address);
    }
    
    if ((address & 0x3fef) == 0x3f00) {
	PPU_bank_3c[0x300] = value;
	PPU_bank_3c[0x304] = value;
	PPU_bank_3c[0x308] = value;
	PPU_bank_3c[0x30c] = value;
	PPU_bank_3c[0x310] = value;
	PPU_bank_3c[0x314] = value;
	PPU_bank_3c[0x318] = value;
	PPU_bank_3c[0x31c] = value;
	PPU_palette_ram[0x00] = vid_pre_xlat[value & 0x3f];
	PPU_palette_ram[0x04] = vid_pre_xlat[value & 0x3f];
	PPU_palette_ram[0x08] = vid_pre_xlat[value & 0x3f];
	PPU_palette_ram[0x0c] = vid_pre_xlat[value & 0x3f];
	PPU_palette_ram[0x10] = vid_pre_xlat[value & 0x3f];
	PPU_palette_ram[0x14] = vid_pre_xlat[value & 0x3f];
	PPU_palette_ram[0x18] = vid_pre_xlat[value & 0x3f];
	PPU_palette_ram[0x1c] = vid_pre_xlat[value & 0x3f];
    } else if ((address & 0x3f00) == 0x3f00) {
	PPU_palette_ram[address & 0x1f] = vid_pre_xlat[value & 0x3f];
	PPU_bank_3c[0x300+(address & 0x1f)] = value;
    } else if ((address > 0x1fff) && (address < 0x3000)) {
	(*PPU_banks[(address >> 10)])[(address & 0x3ff)] = value;
	ppu_flags |= PPUF_DIRTY_NAMES;
    } else if (address > 0x2fff) {
	(*PPU_banks[(address >> 10)])[(address & 0x3ff)] = value;	
    } else if (!(ppu_flags & PPUF_VROM)) {
	(*PPU_banks[(address >> 10)])[(address & 0x3ff)] = value;
	ppu_flags |= PPUF_DIRTY_PATTERNS;
    }
}

unsigned char PPU_read(unsigned short address)
{
    unsigned char value;
    
    address &= 0x3fff;
    if (address >= 0x3f00) {
	address &= 0x3f1f;
    }
    
    value = (*PPU_banks[(address >> 10)])[(address & 0x3ff)];

#ifdef DEBUG_io_PPU_data
    deb_printf("PPU_write(%04hx) = %02hx\n", address, value);
#endif

    return value;
}

#define VPOS(x) (PPU_sprite_ram[((x) << 2) + 0])
#define XPOS(x) (PPU_sprite_ram[((x) << 2) + 3])
#define TILE(x) (PPU_sprite_ram[((x) << 2) + 1])
#define FLAGS(x) (PPU_sprite_ram[((x) << 2) + 2])
#define VFLIP(x) (FLAGS(x) & 0x80)

struct {
    unsigned char hpos;
    unsigned char pattern0;
    unsigned char pattern8;
    unsigned char flags;
} spritecache[16];

#define HPOS(x) (spritecache[(x)].hpos)
#define ATTRS(x) (spritecache[(x)].flags)
#define PRIO(x) (spritecache[(x)].flags & 0x20)
#define COLOR(x) (spritecache[(x)].flags & 3)
#define HFLIP(x) (spritecache[(x)].flags & 0x40)
#define PAT0(x) (spritecache[(x)].pattern0)
#define PAT8(x) (spritecache[(x)].pattern8)

int num_sprites_line;

void PPU_init_spritecache(unsigned short line, int is_8x16)
{
    int i;
    unsigned char cur_tile;
    unsigned char cur_pattern_line;
    unsigned char *cur_pattern;
    int cur_ptbl;
    
    num_sprites_line = 0;

    cur_ptbl = (ppu_cntrl1 & 0x08)? 1: 0;
    
    /* FIXME: clear sprite count */
    
    for (i = 0; i < 64; i++) {
	if ((VPOS(i) + 1 <= line) && (VPOS(i) + (is_8x16? 17: 9) > line)) {
	    if (num_sprites_line > 15) {
		/* FIXME: set sprite count */
	    } else {
		HPOS(num_sprites_line) = XPOS(i);
		ATTRS(num_sprites_line) = FLAGS(i);

		cur_pattern_line = (line - (VPOS(i)+1)) & (is_8x16? 15: 7);
		if (VFLIP(i)) {
		    cur_pattern_line = (is_8x16? 15: 7) - cur_pattern_line;
		}

		cur_tile = TILE(i);
		if (is_8x16) {
		    cur_ptbl = cur_tile & 1;
		    cur_tile = (cur_pattern_line & 8)?
			(cur_tile | 1):	(cur_tile & 0xfe);
		    cur_pattern_line &= 7;
		}

		if (ppu_latchfunc) {
		    ppu_latchfunc((cur_ptbl << 12) | (cur_tile << 4) | (cur_pattern_line));
		    ppu_latchfunc((cur_ptbl << 12) | (cur_tile << 4) | (cur_pattern_line) | 8);
		}
		
		cur_pattern = &(*PPU_patterns[cur_ptbl][(cur_tile >> 6)])
		    [((cur_tile & 0x3f) << 4) + cur_pattern_line];
		
		PAT0(num_sprites_line) = cur_pattern[0];
		PAT8(num_sprites_line) = cur_pattern[8];

		if ((i == 0) && (ppu_phase == P_PRE_HIT) &&
		    ((cur_pattern[0]) || cur_pattern[8])) {
		    ppu_phase = P_POST_HIT;
		    ppu_status |= 0x40;
		}

		num_sprites_line++;
	    }
	}
    }
}

void PPU_render_sprites(unsigned char *cur_vbp, int behind)
{
    int i;
    unsigned char *colors;

    if (num_sprites_line) {
	for (i = (num_sprites_line - 1); i >= 0; i--) {
	    if (behind? PRIO(i): !PRIO(i)) {
		colors = &PPU_palette_ram[16+(COLOR(i) << 2)];
		if (!HFLIP(i)) {
		    blit_2_8(PAT8(i), PAT0(i), 8, 0, colors, cur_vbp + HPOS(i));
		} else {
		    blit_2_8_rev(PAT8(i), PAT0(i), 8, 0, colors, cur_vbp + HPOS(i));
		}
	    }
	}
    }
}

void PPU_render_background(unsigned short line, char *cur_vbp)
{
    unsigned char pattern0[33];
    unsigned char pattern8[33];
    unsigned char attrs[33];
    int i;
    unsigned char *colors;

    {
	unsigned char *cur_tile;
	unsigned char tile_offset;
	unsigned char *cur_bank;
	unsigned char *cur_attrline;
	unsigned char cur_patternline;
	unsigned char cur_attrbase;
	unsigned char ***cur_ptbl;

	cur_patternline = line & 7;
	cur_attrbase = (line & 16) >> 2;
	cur_ptbl = PPU_patterns[(ppu_cntrl1 & 0x10) >> 4];
	
	cur_bank = PPU_nametables[ppu_saved_nametable];
	tile_offset = (ppu_saved_hscrol >> 3) & 0x1f;
	cur_tile = &cur_bank[((line << 2) & 0x3e0) + tile_offset];
	cur_attrline = cur_bank + ((line >> 2) & 0x38) + 0x3c0 + (tile_offset >> 2);

	cur_attrbase |= tile_offset & 2;
	
	for (i = 0; i < 33; i++) {
	    {
		unsigned char *cur_pattern;

		cur_pattern = &(*cur_ptbl[(*cur_tile >> 6)])
		    [((*cur_tile & 0x3f) << 4) + cur_patternline];
		pattern0[i] = cur_pattern[0];
		pattern8[i] = cur_pattern[8];
	    }
	    
	    if (ppu_latchfunc) {
		ppu_latchfunc(((ppu_cntrl1 & 0x10) << 8) | ((*cur_tile) << 4) | cur_patternline);
		ppu_latchfunc(((ppu_cntrl1 & 0x10) << 8) | ((*cur_tile) << 4) | cur_patternline | 8);
	    }
	    
	    attrs[i] = ((*cur_attrline >> (cur_attrbase)) & 3) << 2;

	    if (tile_offset & 1) {
		cur_attrbase ^= 2;
	    }

	    if ((tile_offset & 3) == 3) {
		cur_attrline++;
	    }
	    
	    cur_tile++;
	    tile_offset++;
	    if (tile_offset > 31) {
		tile_offset = 0;
		cur_bank = PPU_nametables[ppu_saved_nametable ^ 1];
		cur_tile = &cur_bank[((line << 2) & 0x3e0)];
		cur_attrline = cur_bank + ((line >> 2) & 0x38) + 0x3c0;
	    }
	}
    }
    
    if (ppu_cntrl2 & 0x02) {
	i = 0;
    } else {
	for (i = 0; i < 8; i++) {
	    *cur_vbp++ = vid_pre_xlat[0x0f];
	}
	i = 1;
    }

    if (ppu_saved_hscrol & 7) {
	colors = &PPU_palette_ram[attrs[i]];

	blit_2_8(pattern8[i], pattern0[i], 8 - (ppu_hscrol & 7), ppu_hscrol & 7, colors, cur_vbp);
	i++;
	cur_vbp += 8 - (ppu_hscrol & 7);
    }
    
    while (i < 32) {
	colors = &PPU_palette_ram[attrs[i]];
	blit_2_8(pattern8[i], pattern0[i], 8, 0, colors, cur_vbp);
	cur_vbp += 8;
	i++;
    }

    if (ppu_saved_hscrol & 7) {
	
	colors = &PPU_palette_ram[attrs[i]];

	blit_2_8(pattern8[i], pattern0[i], ppu_hscrol & 7, 0, colors, cur_vbp);
    }
}

void PPU_render_scan_line(unsigned short line)
{
    unsigned char *cur_vbp;
    
    cur_vbp = video_get_vbp(line);
    
    if ((!(ppu_flags & PPUF_NTSC)) || ((line > 7) && (line < 231))) {
	memset(cur_vbp, PPU_palette_ram[0], 256);
	
	if (ppu_cntrl2 & 0x10) {
	    PPU_init_spritecache(line, ppu_cntrl1 & 0x20);
	    PPU_render_sprites(cur_vbp, 1);
	}
	if (ppu_cntrl2 & 0x08) {
	    PPU_render_background(ppu_saved_vscrol, cur_vbp);
	}
	if (ppu_cntrl2 & 0x10) {
	    PPU_render_sprites(cur_vbp, 0);
	}
    } else {
	memset(cur_vbp, vid_pre_xlat[0x0f], 256);
    }

    if (++ppu_saved_vscrol > 239) {
	ppu_saved_vscrol = 0;
	ppu_saved_nametable ^= 2;
    }
}

void PPU_render_video_frame(void)
{
    video_display_buffer();
}

void PPU_do_dma(int page)
{
    if (page < 0x20) { /* in ram area */
	memcpy(PPU_sprite_ram, bank_0 + ((page & 7) << 8), 0x100);
    } else if (page < 0x60) {
	deb_printf("PPU_do_dma(): sprite DMA from I/O space, ignoring.\n");
    } else if (page < 0x80) {
	memcpy(PPU_sprite_ram, bank_6 + ((page & 0x1f) << 8), 0x100);
    } else if (page < 0xa0) {
	memcpy(PPU_sprite_ram, bank_8 + ((page & 0x1f) << 8), 0x100);
    } else if (page < 0xc0) {
	memcpy(PPU_sprite_ram, bank_A + ((page & 0x1f) << 8), 0x100);
    } else if (page < 0xe0) {
	memcpy(PPU_sprite_ram, bank_C + ((page & 0x1f) << 8), 0x100);
    } else {
	memcpy(PPU_sprite_ram, bank_E + ((page & 0x1f) << 8), 0x100);
    }
}

void ppu_w_2000(unsigned char value)
{
    ppu_cntrl1 = value;
    ppu_saved_nametable = value & 3;
}

void ppu_w_2001(unsigned char value)
{
    ppu_cntrl2 = value;
    if ((value & 0x08) && (ppu_phase != P_VBLANK)) {
	ppu_status |= 0x10;
    }
}

void ppu_w_2002(unsigned char value)
{
    deb_printf("ppu_w_2002(%02x): Unknown memory location.\n", value);
}

void ppu_w_2003(unsigned char value)
{
    ppu_spraddy = value;
}

void ppu_w_2004(unsigned char value)
{
    PPU_sprite_ram[ppu_spraddy++] = value;
}

void ppu_w_2005(unsigned char value)
{
    if (ppu_scrollmode & 1) {
	ppu_scrollmode &= ~1;
	if (value > 239) {
	    ppu_vscrol = 0;
	} else {
	    ppu_vscrol = value;
	}
    } else {
	ppu_scrollmode |= 1;
	ppu_hscrol = value;
	ppu_saved_hscrol = value;
    }
}

void ppu_w_2006(unsigned char value)
{
    if (PPU_amode & 1) {
	PPU_amode &= ~1;
	PPU_addr = (ppu_addrsave << 8) | value;
	/* this next bit is for mid-hblank writes. */
	ppu_saved_nametable = (ppu_addrsave >> 2) & 3;
	ppu_saved_vscrol = ((ppu_addrsave >> 4) & 3) + ((ppu_addrsave & 3) << 6) + ((value >> 2) & 0x38);
	if (ppu_saved_vscrol > 239) {
	    ppu_saved_vscrol -= 239;
	    ppu_saved_nametable ^= 2;
	}
	ppu_saved_hscrol = ((value << 3) & 0xf8);
    } else {
	PPU_amode |= 1;
	ppu_addrsave = value;
    }
}

void ppu_w_2007(unsigned char value)
{
    if (PPU_amode & 1) {
	deb_printf("ppu_w_2007(): ppu write during addr sequence.\n");
    }
    PPU_write(PPU_addr, value);
    PPU_addr += ((ppu_cntrl1 & 0x04)? 0x20: 1);
}

unsigned char ppu_r_2000(void)
{
    deb_printf("ppu_r_2000(): Unknown address.\n");
    return 0;
}

unsigned char ppu_r_2001(void)
{
    deb_printf("ppu_r_2001(): Unknown address.\n");
    return 0;
}

unsigned char ppu_r_2002(void)
{
    unsigned char retval;
    
    retval = ppu_status;
    ppu_status &= 0x7f;
    return retval;
}

unsigned char ppu_r_2003(void)
{
    deb_printf("ppu_r_2003(): Unknown address.\n");
    return 0;
}

unsigned char ppu_r_2004(void)
{
    return PPU_sprite_ram[ppu_spraddy++];
}

unsigned char ppu_r_2005(void)
{
    deb_printf("ppu_r_2005(): Unknown address.\n");
    return 0;
}

unsigned char ppu_r_2006(void)
{
    deb_printf("ppu_r_2006(): Unknown address.\n");
    return 0;
}

unsigned char ppu_r_2007(void)
{
    unsigned char retval;

    retval = ppu_read_latch;

    ppu_read_latch = PPU_read(PPU_addr);
    PPU_addr += ((ppu_cntrl1 & 0x04)? 0x20: 1);
    
    return retval;
}

void PPU_dump_memory(short address)
{
    unsigned short i;

    for (i = address; i < address+0x400; i++) {
        if ((i & 0x1f) == 0) {
	    deb_printf("\n0x%04hx:", i);
	}

	if ((i & 0x3) == 0) {
	    deb_printf(" %02x", ((*PPU_banks[i >> 10])[i & 0x3ff]));
	} else {
	    deb_printf("%02x", ((*PPU_banks[i >> 10])[i & 0x3ff]));
	}
    }
    deb_printf("\n");
}

void PPU_dump_sprite_memory(void)
{
    int i;

    for (i = 0; i < 64; i++) {
	deb_printf("Sprite %d: 0x%02x 0x%02x 0x%02x 0x%02x %s\n", i,
		   PPU_sprite_ram[(i << 2) + 0],
		   PPU_sprite_ram[(i << 2) + 1],
		   PPU_sprite_ram[(i << 2) + 2],
		   PPU_sprite_ram[(i << 2) + 3],
		   (PPU_sprite_ram[(i << 2) + 2] & 0x20)? "Behind": "Before");
    }
}

void PPU_check_sprites(void)
{
    int i;
    int j;
    
    for (i = 0; i < 64; i++) {
	for (j = 0; j < 64; j++) {
	    if ((i != j) &&
		(PPU_sprite_ram[(i << 2) + 0] == PPU_sprite_ram[(j << 2) + 0]) &&
		(PPU_sprite_ram[(i << 2) + 3] == PPU_sprite_ram[(j << 2) + 3])) {
		deb_printf ("sprite position match: %d == %d.\n", i, j);
	    }
	}
    }
}

void PPU_dump_pattern(short address)
{
    char *pattern;
    int i;

    pattern = &((*PPU_banks[address >> 10])[address & 0x3ff]);

    deb_printf("pattern:");
    for (i = 0; i < 16; i++) {
	deb_printf(" %02hx", (*(pattern+i)) & 0xff);
    }
    deb_printf("\n");

    for (i = 0; i < 8; i++) {
	unsigned char x;
	unsigned char y;

	x = (*(pattern+i)) & 0xff;
	y = (*(pattern+i+8)) & 0xff;

	deb_printf("graphic: %s%s%s%s%s%s%s%s\n",
		   (x & 0x80)? (y & 0x80)? "*": "+": (y & 0x80)? "-": " ",
		   (x & 0x40)? (y & 0x40)? "*": "+": (y & 0x40)? "-": " ",
		   (x & 0x20)? (y & 0x20)? "*": "+": (y & 0x20)? "-": " ",
		   (x & 0x10)? (y & 0x10)? "*": "+": (y & 0x10)? "-": " ",
		   (x & 0x08)? (y & 0x08)? "*": "+": (y & 0x08)? "-": " ",
		   (x & 0x04)? (y & 0x04)? "*": "+": (y & 0x04)? "-": " ",
		   (x & 0x02)? (y & 0x02)? "*": "+": (y & 0x02)? "-": " ",
		   (x & 0x01)? (y & 0x01)? "*": "+": (y & 0x01)? "-": " ");
    }
}

/*
 * $Log: nes_ppu.c,v $
 * Revision 1.58  1999/01/11 02:23:17  nyef
 * changed to use blitters from blitters.c instead of nes_blit.c.
 *
 * Revision 1.57  1998/12/15 01:50:51  nyef
 * changed mapper hsync and vsync routines to conform to new spec.
 *
 * Revision 1.56  1998/12/12 03:28:32  nyef
 * changed ppu_r_2002() to clear the VBlank bit after reading ppu_status.
 *
 * Revision 1.55  1998/12/11 04:33:40  nyef
 * began fixing for full compatability with cal.
 *
 * Revision 1.54  1998/12/11 02:50:57  nyef
 * removed mid-hblank write detection logic from ppu_w_2006().
 *
 * Revision 1.53  1998/12/09 04:36:46  nyef
 * put the vblank length back down to what I originally worked it out as.
 * also changed some of the vblank timing signals.
 *
 * Revision 1.52  1998/12/04 02:18:51  nyef
 * in ppu_render_background(), moved call of ppu_latchfunc() to after
 * obtaining the address of the tile pattern. this fixed the right side
 * of the windows in fire emblem (mapper 10).
 *
 * Revision 1.51  1998/11/28 06:32:41  nyef
 * removed some superfluous code from PPU_render_background().
 *
 * Revision 1.50  1998/11/28 05:14:08  nyef
 * changed ppu_render_scan_line() to use memset to clear the video buffer.
 * Last time this was done, one of the clears was missed.
 * Also cleaned out some commented out code in ppu_render_scan_line().
 *
 * Revision 1.49  1998/11/05 17:02:33  nyef
 * performed strength reduction on tile attribute calculation
 * in ppu_render_background().
 *
 * Revision 1.48  1998/11/05 16:55:57  nyef
 * changed ppu_render_background() to obtain nice speedup.
 *
 * Revision 1.47  1998/11/05 15:17:20  nyef
 * simplified (slightly) the invocation of ppu_latchfunc() in
 * ppu_render_background().
 *
 * Revision 1.46  1998/11/05 04:05:05  nyef
 * changed ppu_render_scan_line() to use memset to clear the video buffer.
 *
 * Revision 1.45  1998/11/01 21:28:10  nyef
 * doubled vsync time again. Bucky O' Hare now works properly.
 *
 * Revision 1.44  1998/11/01 21:17:34  nyef
 * re-added the assignment to ppu_saved_nametable in ppu_enter_cpu().
 * also added some debugging code to ppu_w_2007().
 *
 * Revision 1.43  1998/11/01 18:18:35  nyef
 * added support for left-clipped background.
 *
 * Revision 1.42  1998/11/01 14:17:43  nyef
 * moved assignment to ppu_saved_nametable from ppu_enter_cpu()
 * to ppu_w_2000().
 *
 * Revision 1.41  1998/11/01 03:09:45  nyef
 * changed ppu_saved_hscrol, ppu_saved_vscrol, and ppu_saved_nametable
 * from int to unsigned char.
 *
 * Revision 1.40  1998/10/31 04:36:39  nyef
 * removed cur_attr from ppu_render_background.
 *
 * Revision 1.39  1998/10/31 04:02:12  nyef
 * re-added (preliminary) support for ppu reads.
 *
 * Revision 1.38  1998/10/31 03:55:32  nyef
 * removed some useless code from ppu_render_background()
 *
 * Revision 1.37  1998/10/29 03:29:23  nyef
 * fixed minor mistake in ppu_render_background().
 *
 * Revision 1.36  1998/10/29 03:17:34  nyef
 * Added preliminary support for Mid-HBlank PPU address writes.
 * There are still some graphics issues, but Crystalis and Kirby
 * are both playable.
 *
 * Revision 1.35  1998/10/29 01:29:07  nyef
 * moved all PPU I/O access in from io_nes.S.
 * access to $2007 is unimplimented.
 *
 * Revision 1.34  1998/10/22 02:02:59  nyef
 * moved ppu_cntrl1, ppu_cntrl2, ppu_status, ppu_spraddy, ppu_vscrol,
 * ppu_hscrol, ppu_scrollmode, PPU_addr and PPU_amode in from io_nes.S
 *
 * Revision 1.33  1998/10/19 13:07:33  nyef
 * added variable ppu_saved_vscrol to prevent vscroll outside of vblank.
 *
 * Revision 1.32  1998/10/03 20:38:43  nyef
 * changed PPU_write() to be more selective about palette values.
 *
 * Revision 1.31  1998/10/01 02:33:52  nyef
 * added preliminary support for latches as per mappers 9 and 10
 *
 * Revision 1.30  1998/09/26 13:58:14  nyef
 * fixed bug in PPU_dump_sprite_memory. (fixes behind/before display)
 *
 * Revision 1.29  1998/09/20 21:26:23  nyef
 * added two new debug hooks. (PPU_dump_sptite_memory and PPU_check_sprites)
 *
 * Revision 1.28  1998/09/12 19:09:43  nyef
 * added a check in the mirror control code so as to not spam debug
 * messages about setting the mirror state to the current one.
 *
 * Revision 1.27  1998/09/08 01:51:43  nyef
 * cleaned up the right-clip blit case in PPU_render_background.
 * more cleanup of the left-clip blit case in PPU_render_background.
 *
 * Revision 1.26  1998/09/08 01:35:04  nyef
 * cleaned up the left-clip blit cases in PPU_render_background.
 *
 * Revision 1.25  1998/09/08 01:09:08  nyef
 * minor bugfixes in PPU_write. (it wasn't handling palette writes correctly)
 *
 * Revision 1.24  1998/08/31 01:35:14  nyef
 * found that some games work better when PPU_banks is changed with
 * the mirroring.
 *
 * Revision 1.23  1998/08/30 20:36:19  nyef
 * added a call to the psg emulation on every vsync.
 *
 * Revision 1.22  1998/08/30 17:04:50  nyef
 * removed PPU_use_vrom and PPU_ntsc in favor of a more general ppu_flags
 * system. also added flags for nametable and patterntable modifications.
 *
 * Revision 1.21  1998/08/30 16:41:37  nyef
 * added a check of scrollflag in ppu_render_background.
 * this fixes some vscroll problems.
 *
 * Revision 1.20  1998/08/30 16:02:09  nyef
 * changed the mirroring to use a separate set of pointers so as to not
 * mess up the normal PPU memory map.
 *
 * Revision 1.19  1998/08/30 03:45:29  nyef
 * cleaned up some stuff left from the removal of the video buffer.
 * made a change to the way palette writes are handled in PPU_write().
 *
 * Revision 1.18  1998/08/29 22:10:33  nyef
 * moved control of all video buffers to the video interface.
 *
 * Revision 1.17  1998/08/28 03:06:16  nyef
 * changed PPU_write and the renderers to do some of the color translation.
 *
 * Revision 1.16  1998/08/26 02:38:20  nyef
 * added an additional indirection layer around the pattern tables.
 * used this to slightly clean up the renderers.
 *
 * Revision 1.15  1998/08/26 01:33:12  nyef
 * changed sprite dma to use memcpy(3)s instead of a for loop.
 * (loop invariant code motion)
 *
 * Revision 1.14  1998/08/23 22:42:22  nyef
 * added hooks for calling mappers on vsync and hsync.
 *
 * Revision 1.13  1998/08/16 20:42:58  nyef
 * clarified (and sped up) the behavior of the first for loop in
 * PPU_render_background(). (It still needs more work).
 *
 * Revision 1.12  1998/08/16 12:05:47  nyef
 * fixed bug with VFlipped 8x16 sprites.
 *
 * Revision 1.11  1998/08/15 17:30:43  nyef
 * added 8x16 sprite support.
 *
 * Revision 1.10  1998/08/15 15:36:29  nyef
 * changed PPU_render_sprites() to take advantage of the new reversed
 *   blitters. also clarified the behavior of the renderer.
 *
 * Revision 1.9  1998/08/15 05:00:04  nyef
 * changed PPU_render_background to take advantage of the new
 *   left-side blitters
 *
 * Revision 1.8  1998/08/15 04:16:20  nyef
 * changed PPU_render_background to take advantage of the new
 *   right-side blitters
 *
 * Revision 1.7  1998/08/02 13:37:31  nyef
 * changed every printf to deb_printf for debug console.
 *
 * Revision 1.6  1998/07/25 16:56:29  nyef
 * changed PPU_render_background to use the new blit compiler.
 *
 * Revision 1.5  1998/07/22 02:42:51  nyef
 * added default case to switch statement in PPU_render_background(). oops.
 * clarified (and sped up) the behavior of the first for loop in PPU_render_background(). (It still doesn't make as much sense as it should)
 *
 * Revision 1.4  1998/07/19 04:02:21  nyef
 * fixed the hscroll bug in blaster master (hopefully for the last time)
 *
 * Revision 1.3  1998/07/19 03:32:37  nyef
 * fixed sprite priority by inverting the draw order
 *
 * Revision 1.2  1998/07/18 21:43:17  nyef
 * added support for one-screen mirroring
 *
 * Revision 1.1  1998/07/11 22:18:17  nyef
 * Initial revision
 *
 */
