/*
 * pce.c
 *
 * contains PC-Engine specific code and data.
 */

/* $Id: pce.c,v 1.27 1999/02/20 21:28:03 nyef Exp $ */

#include <stdlib.h>
#include <stdio.h>

#include "emu6280.h"
#include "tool.h"
#include "system.h"
#include "debug.h"
#include "cal.h"
#include "ui.h"
#include "pce_vdp.h"

cal_cpu pce_cpu;
pce_vdp pce_video;
int pce_joyswitch;

#define PCE_RAMSIZE 0x8000

unsigned char *pce_mmap[256];

extern unsigned char joy1_data;

static const unsigned char pce_joydir_xlate[16] = {
    0x0, 0x1, 0x4, 0x5,
    0x8, 0x9, 0xc, 0xd,
    0x2, 0x3, 0x6, 0x7,
    0xa, 0xb, 0xe, 0xf,
};

unsigned char pce_read_mem(cal_cpu cpu, unsigned long addr)
{
    if (cpu->data.d_emu6280->mpr[(addr >> 13) & 7] == 0xff) {
	switch (addr & 0x1c00) {
	case 0x0000: /* HuC6270 VDC */
	    return pce_video->readvdp(pce_video, addr);
	    break;
	case 0x0400: /* HuC6260 VCE */
	    return pce_video->readvce(pce_video, addr);
	    break;
	case 0x0800: /* PSG */
	    break;
	case 0x0c00: /* Timers */
	    break;
	case 0x1000: /* I/O (joypad) */
	    if (pce_joyswitch) {
		return pce_joydir_xlate[(joy1_data >> 4)] ^ 0xff;
	    } else {
		return (joy1_data & 0x0f) ^ 0xff;
	    }
	    break;
	case 0x1400: /* IRQ control */
	    break;
	default:
	    deb_printf("pce: I/O read 0x%04hx, returning 0x00.\n", addr & 0x1fff);
	}
	return 0x00;
    } else {
	return pce_mmap[cpu->data.d_emu6280->mpr[(addr >> 13) & 7]][addr & 0x1fff];
    }
}

void pce_write_mem(cal_cpu cpu, unsigned long addr, unsigned char data)
{
    if (cpu->data.d_emu6280->mpr[(addr >> 13) & 7] == 0xff) {
	switch (addr & 0x1c00) {
	case 0x0000: /* HuC6270 VDC */
	    pce_video->writevdp(pce_video, addr, data);
	    break;
	case 0x0400: /* HuC6260 VCE */
	    pce_video->writevce(pce_video, addr, data);
	    break;
	case 0x0800: /* PSG */
	    break;
	case 0x0c00: /* Timers */
	    break;
	case 0x1000: /* I/O (joypad) */
	    pce_joyswitch = data & 1;
	    break;
	case 0x1400: /* IRQ control */
	    break;
	default:
	    deb_printf("pce: I/O write 0x%04hx = 0x%02x.\n", addr & 0x1fff, data);
	}
    } else if (cpu->data.d_emu6280->mpr[(addr >> 13) & 7] == 0xf7) {
	/* Savegame RAM */
	pce_mmap[cpu->data.d_emu6280->mpr[(addr >> 13) & 7]][addr & 0x1fff] = data;
    } else if (cpu->data.d_emu6280->mpr[(addr >> 13) & 7] > 0xf7) {
	/* Normal RAM */
	pce_mmap[cpu->data.d_emu6280->mpr[(addr >> 13) & 7]][addr & 0x1fff] = data;
    } else {
	/* Ignore it, it's in the ROM area */
    }
}

memread_t pce_memread[1] = {pce_read_mem,};
memwrite_t pce_memwrite[1] = {pce_write_mem,};

void pce_st0_hack(unsigned char data)
{
    pce_video->writevdp(pce_video, 0, data);
}

void pce_st1_hack(unsigned char data)
{
    pce_video->writevdp(pce_video, 2, data);
}

void pce_st2_hack(unsigned char data)
{
    pce_video->writevdp(pce_video, 3, data);
}

void pce_timeslice(cal_cpu cpu)
{
    int i;

    for (i = 243; i && (system_flags == F_NONE); i--) {
	cpu->runfor(cpu, 230);
	cpu->run(cpu);
	if (pce_video->periodic(pce_video)) {
/* 	    deb_printf("pce_video->periodic() requests interrupt.\n"); */
	    cpu->irq(cpu);
	}
	if (system_flags & F_UNIMPL) {
/* 	    deb_printf("Unimplimented Instruction.\n"); */
	    deb_printf("Unimplimented Instruction 0x%02x.\n", pce_read_mem(cpu, cpu->data.d_emu6280->reg_PC));
	}
    }
}

void pce_init_mmap(void)
{
    int i;
    int offset;

    offset = 0;
    
    if (romsize & 0x200) {
	printf("pce_init_mmap(): compensating for 512-byte header.\n");
	offset = 0x200;
    }
    
    for (i = 0; i < 0xf7; i++) {
	pce_mmap[i] = romimage +
	    ((((i * 0x2000) + offset) < romsize)? ((i * 0x2000) + offset): 0);
    }

    pce_mmap[0xf7] = NULL;

    for (i = 0xf8; i < 0xfc; i++) {
	pce_mmap[i] = ramblock + ((i - 0xf8) + 0x2000);
    }

    for (i = 0xfc; i < 0x100; i++) {
	pce_mmap[i] = NULL;
    }
}

void pce_run(int startup_deb)
{
    ramblock = malloc(PCE_RAMSIZE);

    pce_init_mmap();
    pce_cpu = cal_create(CPUT_EMU6280);

    pce_video = pce_vdp_create();
    
    if ((romimage != NULL) && (ramblock != NULL) && (pce_cpu) && (pce_video)) {
	pce_cpu->setzpage(pce_cpu, ramblock);
	pce_cpu->setmmu(pce_cpu, 0, 0, pce_memread, pce_memwrite);
	pce_cpu->reset(pce_cpu);
	
	set_timeslice((void (*)(void *)) &pce_timeslice, pce_cpu);
    } else {
	printf("out of memory\n");
    }
}

/*
 * $Log: pce.c,v $
 * Revision 1.27  1999/02/20 21:28:03  nyef
 * doubled number of cycles per scanline
 *
 * Revision 1.26  1999/02/15 03:25:52  nyef
 * added patch from Michael Vance to squash warnings
 *
 * Revision 1.25  1999/02/03 02:43:12  nyef
 * added translation table to fix joypad directions
 *
 * Revision 1.24  1999/02/02 16:20:34  nyef
 * added preliminary support for joypad 1
 *
 * Revision 1.23  1999/01/31 21:16:09  nyef
 * added debug output of joypad access
 *
 * Revision 1.22  1999/01/31 21:12:09  nyef
 * re-enabled interrupt generation
 *
 * Revision 1.21  1999/01/31 20:26:39  nyef
 * disabled interrupt generation.
 *
 * Revision 1.20  1999/01/30 23:27:30  nyef
 * added support for the pce vce
 *
 * Revision 1.19  1999/01/30 23:09:50  nyef
 * redesigned VDP I/O interface.
 *
 * Revision 1.18  1999/01/30 19:58:29  nyef
 * added i/o cases for all major devices.
 *
 * Revision 1.17  1999/01/30 19:30:13  nyef
 * reenabled output of I/O access to unknown ports.
 * added support for writing to RAM areas.
 *
 * Revision 1.16  1999/01/28 02:17:24  nyef
 * added routines to support the ST0, ST1, and ST2 instructions.
 *
 * Revision 1.15  1999/01/26 03:05:42  nyef
 * fixed to use 32bit "address bus"
 *
 * Revision 1.14  1999/01/17 02:12:43  nyef
 * added support for the new pce vdp
 *
 * Revision 1.13  1999/01/13 03:55:36  nyef
 * fixed to use the cal, the timeslicer, and the new memory access spec.
 *
 * Revision 1.12  1998/07/27 00:57:40  nyef
 * added compensation for 512-byte header in pce_init_mmap()
 *
 * Revision 1.11  1998/07/26 23:11:56  nyef
 * changed read/write table allocation to allocate extra room for zpage
 *
 * Revision 1.10  1998/07/14 10:25:21  nyef
 * added temporary assignments to cycles_left in pce_trace and pce_freerun
 *
 * Revision 1.9  1998/07/14 01:52:38  nyef
 * added initializer for disassemble()
 *
 * Revision 1.8  1998/07/14 01:37:21  nyef
 * added trace function for debugger
 *
 * Revision 1.7  1998/07/12 23:22:00  nyef
 * fixed two stupid mistakes in the mmap initializer
 *
 * Revision 1.6  1998/07/12 23:21:03  nyef
 * added initializer for memory mapping
 *
 * Revision 1.5  1998/07/12 23:03:07  nyef
 * fixed mem_read_byte()
 *
 * Revision 1.4  1998/07/12 22:36:05  nyef
 * set ram size to 32k (this figure is enough to cover the SuperGrafx)
 *
 * Revision 1.3  1998/07/12 21:50:00  nyef
 * finished fixing display_vectors()
 *
 * Revision 1.2  1998/07/12 21:24:57  nyef
 * started fixing display_vectors()
 * added freerun()
 * made regs static
 *
 * Revision 1.1  1998/07/11 22:17:59  nyef
 * Initial revision
 *
 */
